# -*- coding: utf-8 -*-
# Copyright (C) 2019 fjord-technologies
# SPDX-License-Identifier: GPL-3.0-or-later
"""xtrargon_ng.modules.server"""

import copy
import logging
import os
import subprocess
import netifaces

from dwho.classes.modules import MODULES
from dwho.classes.notifiers import DWhoPushNotifications
from httpdis.ext.httpdis_json import HttpReqErrJson
from sonicprobe import helpers
from sonicprobe.libs import xys
from sonicprobe.libs.moresynchro import RWLock

from ..classes.abstract.modules import XtrargonDBModuleBase


DEFAULT_KA_FORBID_VIP_FILE = "/keepalived_forbid_vip"
DEFAULT_KA_FORCE_CK_STATUS = "/keepalived_force_check_status"
DEFAULT_MAINT_FILE         = "/var/tmp/fd-maint.yes"

KA_IS_ACTIVE_CMD           = ('systemctl', 'is-active', 'keepalived.service')

DRBD_STATE_FILE            = "/proc/drbd"

LOG                        = logging.getLogger('xtrargon_ng.modules.server')


class ServerModule(XtrargonDBModuleBase):
    MODULE_NAME = 'server'

    LOCK = RWLock()

    # pylint: disable-msg=attribute-defined-outside-init
    def safe_init(self, options):
        self.lock_timeout = self.config['general']['lock_timeout']

        if not isinstance(self.config.get('db'), dict):
            self.config['db'] = {}

        if 'alive' not in self.config['db']:
            self.config['db']['alive'] = True

        if self.config['db']['alive']:
            self.init_db()

        self.ka_forbid_vip_file = DEFAULT_KA_FORBID_VIP_FILE
        self.ka_force_ck_status_file = DEFAULT_KA_FORCE_CK_STATUS

        if self.config['general'].get('keepalived_forbid_vip'):
            self.ka_forbid_vip_file = self.config['general']['keepalived_forbid_vip']

        if self.config['general'].get('keepalived_force_check_status'):
            self.ka_force_ck_status_file = self.config['general']['keepalived_force_check_status']

        self.notifiers = None
        if self.config['general'].get('notifiers_path'):
            self.notifiers = DWhoPushNotifications(self.server_id,
                                                   self.config['general']['notifiers_path'])

    @staticmethod
    def _touch_maint_file():
        helpers.touch(DEFAULT_MAINT_FILE)

    @staticmethod
    def _exists_maint_file():
        return os.path.exists(DEFAULT_MAINT_FILE)

    @staticmethod
    def _unlink_maint_file():
        xfile = DEFAULT_MAINT_FILE
        if os.path.exists(xfile):
            os.unlink(xfile)

    def has_notifications(self):
        if self.config['db'].get('notifications') \
           and self.notifiers:
            return True

        return False

    def is_primary(self, request): # pylint: disable-msg=unused-argument
        if not self.LOCK.acquire_read(self.lock_timeout):
            raise HttpReqErrJson(503, "unable to take LOCK for reading after %s seconds" % self.lock_timeout)

        try:
            ifname = self.config['general'].get('primary_ifname')

            if not ifname or not helpers.has_len(ifname):
                raise HttpReqErrJson(500, "missing primary_ifname option in section general")

            interfaces = netifaces.interfaces() # pylint: disable-msg=c-extension-no-member

            if ifname in interfaces:
                return True

            if ':' not in ifname:
                return False

            ifname = ifname.rstrip('0123456789')

            for interface in interfaces:
                if interface.startswith(ifname):
                    return True

            return False
        except HttpReqErrJson:
            raise
        except Exception as e:
            LOG.exception(e)
            raise HttpReqErrJson(503, repr(e))
        finally:
            self.LOCK.release()

    def has_drbd(self, request): # pylint: disable-msg=unused-argument
        if not self.LOCK.acquire_read(self.lock_timeout):
            raise HttpReqErrJson(503, "unable to take LOCK for reading after %s seconds" % self.lock_timeout)

        try:
            return os.path.isfile(DRBD_STATE_FILE)
        except HttpReqErrJson:
            raise
        except Exception as e:
            LOG.exception(e)
            raise HttpReqErrJson(503, repr(e))
        finally:
            self.LOCK.release()

    def forbid_vip(self, request): # pylint: disable-msg=unused-argument
        if not self.LOCK.acquire_read(self.lock_timeout):
            raise HttpReqErrJson(503, "unable to take LOCK for reading after %s seconds" % self.lock_timeout)

        f = None

        try:
            LOG.info("check if file %r exists", self.ka_forbid_vip_file)
            if os.path.isfile(self.ka_forbid_vip_file):
                return True

            LOG.info("check if file %r exists and contents equal '1'", self.ka_force_ck_status_file)
            if os.path.isfile(self.ka_force_ck_status_file):
                with open(self.ka_force_ck_status_file, 'r') as f:
                    if f.read(2).strip() == '1':
                        return True

            LOG.info("check if keepalived is active")
            rc = subprocess.call(KA_IS_ACTIVE_CMD,
                                 stdout = subprocess.PIPE,
                                 stderr = subprocess.PIPE)
            return int(rc) != 0
        except HttpReqErrJson:
            raise
        except Exception as e:
            LOG.exception(e)
            raise HttpReqErrJson(503, repr(e))
        finally:
            if f:
                f.close()
            self.LOCK.release()


    LOCK_DB_QSCHEMA = xys.load("""
    read_only*: !!str
    """)

    def do_lock_db(self, request):
        params = request.query_params()

        if not isinstance(params, dict):
            raise HttpReqErrJson(400, "invalid arguments type")

        if not xys.validate(params, self.LOCK_DB_QSCHEMA):
            raise HttpReqErrJson(415, "invalid arguments for command")

        if not self.LOCK.acquire_write(self.lock_timeout):
            raise HttpReqErrJson(503, "unable to take LOCK for writing after %s seconds" % self.lock_timeout)

        try:
            if not self.dbconf or not self.dburi:
                raise HttpReqErrJson(400, "database is configured as not alive")

            if self.lock_db and self.lock_db.is_set():
                raise HttpReqErrJson(400, "database is already locked")

            cfg = copy.deepcopy(self.config)

            if 'read_only' in params:
                if params['read_only'] is None:
                    cfg['db']['read_only'] = None
                elif helpers.boolize(params['read_only']):
                    cfg['db']['read_only'] = True
                else:
                    cfg['db']['read_only'] = params['read_only']

            self._init_lock_db(cfg['db'])
            self._clear_lock_db()
            self._start_lock_db()

            if self.lock_db:
                self.lock_db.set()

            return True
        except HttpReqErrJson:
            raise
        except Exception as e:
            LOG.exception(e)
            raise HttpReqErrJson(503, repr(e))
        finally:
            self.LOCK.release()

    def get_db_lock_status(self, request): # pylint: disable-msg=unused-argument
        if not self.LOCK.acquire_read(self.lock_timeout):
            raise HttpReqErrJson(503, "unable to take LOCK for reading after %s seconds" % self.lock_timeout)

        try:
            if not self.dbconf or not self.dburi:
                raise HttpReqErrJson(400, "database is configured as not alive")

            if not self.lock_db or not self.db_locked:
                return 'uninitialized'

            if self.lock_db.is_set():
                if self.db_locked.is_set():
                    return 'locked'
                return 'locking'

            if self.db_locked.is_set():
                return 'unlocking'
            return 'unlocked'
        except HttpReqErrJson:
            raise
        except Exception as e:
            LOG.exception(e)
            raise HttpReqErrJson(503, repr(e))
        finally:
            self.LOCK.release()

    def do_unlock_db(self, request): # pylint: disable-msg=unused-argument
        if not self.LOCK.acquire_write(self.lock_timeout):
            raise HttpReqErrJson(503, "unable to take LOCK for writing after %s seconds" % self.lock_timeout)

        try:
            if not self.dbconf or not self.dburi:
                raise HttpReqErrJson(400, "database is configured as not alive")

            if not self.lock_db or not self.db_locked:
                raise HttpReqErrJson(400, "database lock isn't initialized")

            if not self.lock_db.is_set() and not self.db_locked.is_set():
                raise HttpReqErrJson(400, "database isn't locked")

            self._clear_lock_db()

            return True
        except HttpReqErrJson:
            raise
        except Exception as e:
            LOG.exception(e)
            raise HttpReqErrJson(503, repr(e))
        finally:
            self.LOCK.release()


    def maint(self, request): # pylint: disable-msg=unused-argument
        if request.get_method() == 'GET':
            if not self.LOCK.acquire_read(self.lock_timeout):
                raise HttpReqErrJson(503, "unable to take LOCK for writing after %s seconds" % self.lock_timeout)
        else:
            if not self.LOCK.acquire_write(self.lock_timeout):
                raise HttpReqErrJson(503, "unable to take LOCK for writing after %s seconds" % self.lock_timeout)

        r = True

        try:
            if request.get_method() == 'DELETE':
                LOG.info("unlink maint file")
                self._unlink_maint_file()
            elif request.get_method() == 'GET':
                LOG.info("check maint file exists")
                r = self._exists_maint_file()
            elif request.get_method() == 'POST':
                LOG.info("write maint file")
                self._touch_maint_file()
            else:
                raise HttpReqErrJson(400, "invalid request method")

            return r
        except HttpReqErrJson:
            raise
        except Exception as e:
            LOG.exception(e)
            raise HttpReqErrJson(503, repr(e))
        finally:
            self.LOCK.release()


if __name__ != "__main__":
    def _start():
        MODULES.register(ServerModule())
    _start()
